#
# field.rb
#
#   Copyright (c) 1998-2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'sync'
require 'delegate'
require 'amstd/to_s'
require 'tmail/mailp'
require 'tmail/encode'


module TMail


  MSGID = /<[^\@>]+\@[^>\@]+>/o

  ZONESTR_TABLE = {
    'jst' =>   9 * 60,
    'eet' =>   2 * 60,
    'bst' =>   1 * 60,
    'met' =>   1 * 60,
    'gmt' =>   0,
    'utc' =>   0,
    'ut'  =>   0,
    'nst' => -(3 * 60 + 30),
    'ast' =>  -4 * 60,
    'edt' =>  -4 * 60,
    'est' =>  -5 * 60,
    'cdt' =>  -5 * 60,
    'cst' =>  -6 * 60,
    'mdt' =>  -6 * 60,
    'mst' =>  -7 * 60,
    'pdt' =>  -7 * 60,
    'pst' =>  -8 * 60,
    'a'   =>  -1 * 60,
    'b'   =>  -2 * 60,
    'c'   =>  -3 * 60,
    'd'   =>  -4 * 60,
    'e'   =>  -5 * 60,
    'f'   =>  -6 * 60,
    'g'   =>  -7 * 60,
    'h'   =>  -8 * 60,
    'i'   =>  -9 * 60,
    # j not use
    'k'   => -10 * 60,
    'l'   => -11 * 60,
    'm'   => -12 * 60,
    'n'   =>   1 * 60,
    'o'   =>   2 * 60,
    'p'   =>   3 * 60,
    'q'   =>   4 * 60,
    'r'   =>   5 * 60,
    's'   =>   6 * 60,
    't'   =>   7 * 60,
    'u'   =>   8 * 60,
    'v'   =>   9 * 60,
    'w'   =>  10 * 60,
    'x'   =>  11 * 60,
    'y'   =>  12 * 60,
    'z'   =>   0 * 60
  }

  TSPECIAL     = %_()<>@,.;:\\"/[]?=_
  CONTROL      = "\\\x00-\\\x19\\\x7f-\\\xff"
  INSECURE     = /[#{Regexp.quote TSPECIAL}#{CONTROL} ]/o
  INSECURE_PH  = /[#{Regexp.quote TSPECIAL}#{CONTROL}]/o

  ENCODED = /=\?[^\s?=]+\?[qb]\?[^\s?=]+\?=/i

  class << self

  def msgid?( str )
    MSGID === str
  end

  def zonestr2i( str )
    if /([\+\-])(\d\d?)(\d\d)/ === str then
      minu = $2.to_i * 60 + $3.to_i
      $1 == '-' ? -minu : minu
    else
      unless tmp = ZONESTR_TABLE[ str.downcase ] then
        raise ParseError, "wrong timezone format '#{str}'"
      end
      tmp * 60
    end
  end

  def time2str( time )
    ret = time.strftime( "%a, #{time.mday} %b %Y %X " )

    # Time#gmtime changes self!!!
    tm = Time.at( time.to_i )
    # [ruby-list:7928]
    offset = tm.to_i - Time.mktime( *tm.gmtime.to_a[0, 6].reverse ).to_i
    ret << '%+.2d%.2d' % (offset / 60).divmod( 60 )

    ret
  end

  def quote( str )
    if INSECURE === str then
      '"' + str.gsub('"', '\\\\"') << '"'
    else
      str
    end
  end

  def quote_phrase( str )
    if INSECURE_PH === str then
      '"' + str.gsub('"', '\\\\"') << '"'
    else
      str
    end
  end

  def encoded?( str )
    ENCODED === str
  end


  # internal use

  def define_header_to_s( mod, use_limit )
    mod.module_eval %^
      def encoded( eol = "\r\n", charset = 'j',
                   ret = ''#{use_limit ? ', limit = 72' : ''} )
        stra = ::TMail::HFencoder.new( ret, eol, charset,
                 #{use_limit ? 'limit' : 'nil'}  )
        accept stra
        stra.write_in
        ret
      end
      alias to_s encoded

      def decoded( eol = "\n", charset = 'e',
                   ret = ''#{use_limit ? ', limit = 72' : ''} )
        stra = ::TMail::HFdecoder.new( ret, charset )
        accept stra
        stra.write_in
        ret
      end
      alias inspect decoded
    ^
  end

  end   # class << self


class HeaderField

  def initialize( fname, fbody, strict )
    @name = fname
    fbody.strip!
    @body = fbody
    @strict = strict

    @illegal = false
    @parsed = false
    @parsing = false
    # @sync = Sync.new
  end


  R     = (1 << 0)  # readable
  W     = (1 << 1)  # writable
  NODUP = (1 << 2)  # dup?
  NOEQL = (1 << 3)  # used by eql? ?

  class << self

    alias hf_original_new new

    def new( fname, fbody, strict = false )
      if self == HeaderField then
        tmp = fname.downcase
        klass = if /\Ax-/ === tmp
                then StringH
                else STR2CLASS[tmp] || UnknownH
                end

        klass.hf_original_new( fname, fbody, strict )
      else
        hf_original_new( fname, fbody, strict )
      end
    end


    def parse_on_rw
      @attrs = nil
    end

    def parse_on_create
      @attrs = nil
    end

    def header_attr( name, type, flags, ali = nil )
      name = _name2str( name )
      type = _type2str( type )
      r = (flags & R != 0)
      w = (flags & W != 0)
      c = (flags & NODUP != 0)
      a = (flags & NOEQL != 0)

      unless a then
        if @attrs then
          @attrs.push name
        else
          @attrs = [name]
        end
      end

      if r then
        module_eval %-
          def #{name}
            ensure_parsed
            @#{name}
          end
        -
      end

      if w then
        module_eval %-
          def #{name}=( arg )
            ensure_parsed
            @#{name} = arg
          end
        -
      end

      if ali then
        ali.each do |a|
          na = _name2str( a )
          module_eval %(alias #{na} #{name})
          module_eval %(alias #{na}= #{name}=) if w
        end
      end
    end

    def redefine_eql
      bug! unless @attrs

      eqls = @attrs.collect {|at| sprintf( '%s.eql?(other.%s)', at, at ) }
      module_eval %-
        def eql?( other )
          super and
          #{eqls.join(' and ')}
        end
      -

      eqls = @attrs.collect {|at| sprintf( '%s == other.%s', at, at ) }
      module_eval %-
        def ==( other )
          super and #{eqls.join(' and ')}
        end
      -
    end

    def empty_able
      bug! unless @attrs

      module_eval %-
        def empty?
          ensure_parsed
          if @illegal then
            /\A\s*\z/ === @body
          else
            not ( #{@attrs.join(' or ')} )
          end
        end
      -
    end
        
  end


  def hash
    @name.hash
  end

  def eql?( oth )
    @name.downcase == oth.name.downcase
  end

  alias == eql?


  def illegal?
    @illegal
  end

  def empty?
    ensure_parsed
    @illegal and (/\A\s*\z/ === @body)
  end


  def name
    @name.dup
  end

  def body
    ensure_parsed
    v = HFdecoder.new( '' )
    do_accept v
    v.write_in
  end

  def body=( str )
    @body = str
    @parsed = false
    @illegal = false
  end

  def accept( strategy )
    ensure_parsed
    strategy.header_name @name
    do_accept strategy
  end

  ::TMail.define_header_to_s self, true


  private

  def ensure_parsed
    begin
      Thread.critical = true   #@sync.lock :EX
      unless @parsed or @parsing then
        @parsing = true
        Thread.critical = false
        parse
        Thread.critical = true
        @parsing = false
        @parsed = true
      end
    ensure
      Thread.critical = false   # @sync.unlock
    end
  end

  # abstract parse

  # abstract do_accept

end



# string type ---------------------------------------------

module NoAttrHeader

      def append_features( mod ) super; mod.
  
  header_attr :body, String, R | W    ; mod.
  redefine_eql                          end

  private

  def do_accept( stra )
    stra.text @body
  end

  def parse
    init_real
    TMail::HFdecoder.decode( @body )
  end

end


class StringH < HeaderField

  parse_on_rw

  include NoAttrHeader

  private

  def init_real
  end

end


# struct type -----------------------------------------

class StructH < HeaderField

  parse_on_rw

  header_attr :comments, Array, R | NODUP | NOEQL


  private

  def init_real
    @comments = []
    init
  end

  def parse
    pre = nil

    begin
      init_real
      Mailp.parse( @body, self, nil )
    rescue ParseError, ScanError
      if not pre and ::TMail.encoded?( @body ) then
        pre = @body
        @body = HFdecoder.decode( pre )
        retry
      elsif pre then
        @body = pre
      end

      @illegal = true
      raise if @strict
    end
  end

end


# unknown type ----------------------------------------

class UnknownH < StructH

  include NoAttrHeader

  private

  def init
  end

end


# date type -------------------------------------------

class DateH < StructH

  parse_on_rw

  header_attr :date, Time, R | W | NODUP

  def date=( arg )
    @date = arg.localtime
  end

  redefine_eql
  empty_able


  private

  def init
    @date = nil
  end

  def do_accept( strategy )
    strategy.meta ::TMail.time2str( @date )
  end

end


# return path type -------------------------------------

class RetpathH < StructH

  parse_on_rw

  header_attr :route, Array,  R | NODUP, [:routes]
  header_attr :addr,  String, R | W

  redefine_eql
  empty_able


  private

  def init
    @route = []
    @addr = nil
  end

  def do_accept( stra )
    stra.meta '<'
    unless route.empty? then
      first = true
      @route.each do |i|
        s = '@' + i
        if first then
          first = false
        else
          s << ','
        end
        stra.meta
      end
      stra.meta ':'
    end
    stra.meta @addr
    stra.meta '>'
  end

end


# address classes ---------------------------------------------


class Address

  def initialize( local, domain )
    @local = local
    @domain = domain
    @phrase = nil
    @route = []
  end

  attr :phrase, true

  attr :route
  alias routes route
  attr_writer :route   # internal use only


  def Address.join( arr )
    arr.collect {|i| ::TMail.quote i }.join('.')
  end

  def local
    @local.collect {|i| ::TMail.quote i }.join('.')
  end

  def domain
    if @domain then
      @domain.collect {|i| ::TMail.quote i }.join('.')
    else
      nil
    end
  end

  def address
    s = local
    if d = domain then
      s << '@' << d
    end

    s
  end

  def address=( str )
    tmp = str.split( '@', 2 )
    @local  = tmp[0].split('.')
    @domain = tmp[1].split('.')
  end

  alias spec address
  alias spec= address=
  alias addr address
  alias addr= address=


  def eql?( other )
    self.type === other        and
    addr      ==  other.addr   and
    route     ==  other.route  and
    phrase    ==  other.phrase
  end
  alias == eql?

  def dup
    i = nil
    n = type.new( @local, @domain )
    n.phrase = @phrase.dup if @phrase
    n.route = @routes.collect{|i| i.dup } unless @routes.empty?
    n
  end

  def accept( stra )
    spec_p = !@phrase and @route.empty?

    if @phrase then
      stra.phrase @phrase
      stra.spc
    end
    tmp = spec_p ? '' : '<'
    unless @route.empty? then
      tmp << @route.collect {|i| '@' + i }.join(',') << ':'
    end
    tmp << address
    tmp << '>' unless spec_p
    stra.meta tmp
  end

  ::TMail.define_header_to_s self, false

end



class AddressGroup < DelegateClass( Array )

  def initialize( name, arg = nil )
    @name = name
    super arg ? arg.dup : []
  end


  attr :name, true
  
  def eql?( oth )
    super( oth ) and self.name == oth.name
  end
  alias == eql?

  def each_address
    bare_each do |mbox|
      if AddressGroup === mbox then
        mbox.each {|i| yield i }
      else
        yield mbox
      end
    end
  end

  def accept( stra )
    stra.phrase @name
    stra.meta ':'
    stra.spc
    first = true
    bare_each do |mbox|
      if first then
        first = false
      else
        stra.meta ','
      end
      stra.spc
      mbox.accept stra
    end
    stra.meta ';'
  end

  ::TMail.define_header_to_s self, false

end


# saddr type -------------------------------------------

class SaddrH < StructH
  
  parse_on_rw

  header_attr :addr, Address, R | W

  redefine_eql
  empty_able


  private

  def init
    @addr = nil
  end

  def do_accept( stra )
    @addr.accept stra
    unless @comments.empty? then
      @comments.each do |c|
        stra.spc
        stra.meta '('
        stra.text c
        stra.meta ')'
      end
    end
  end

end

class SmboxH < SaddrH
end


# maddr type -----------------------------------------

class MaddrH < StructH

  parse_on_rw

  header_attr :addrs, Array, R | NODUP

  redefine_eql
  empty_able


  private

  def init
    @addrs = []
  end

  def do_accept( stra )
    first = true
    @addrs.each do |a|
      if first then
        first = false
      else
        stra.meta ','
        stra.spc
      end
      a.accept stra
    end
    unless @comments.empty? then
      @comments.each do |c|
        stra.spc
        stra.meta '('
        stra.text c
        stra.meta ')'
      end
    end
  end

end

class MmboxH < MaddrH

=begin
  def do_accept( stra )
    first = true
    last = @addrs[-1]
  end
=end

end


# ref type -------------------------------------------

class RefH < StructH

  parse_on_rw

  header_attr :refs, Array, R | NODUP

  redefine_eql
  empty_able


  def each_msgid
    refs.each do |i|
      yield i if ::TMail.msgid? i
    end
  end

  def each_phrase
    refs.each do |i|
      yield i unless ::TMail.msgid? i
    end
  end


  private

  def init
    @refs = []
  end

  def do_accept( stra )
    first = true
    refs.each do |i|
      if first then
        first = false
      else
        stra.spc
      end
      if ::TMail.msgid? i then
        stra.meta i
      else
        stra.phrase i
      end
    end
  end

end


# key type -------------------------------------------

class KeyH < StructH

  parse_on_rw

  header_attr :keys, Array, R | NODUP

  redefine_eql
  empty_able


  private

  def init
    @keys = []
  end

  def do_accept( stra )
    first = true
    keys.each do |i|
      if first then
        first = false
      else
        stra.meta ','
      end
      stra.meta i
    end
  end

end


# received type ---------------------------------------

class RecvH < StructH

  parse_on_rw

  header_attr :from,  String, R | W
  header_attr :by,    String, R | W
  header_attr :via,   String, R | W
  header_attr :with,  Array,  R | NODUP
  header_attr :msgid, String, R | W
  header_attr :for_,  String, R | W,        [:ford, :for_domain]
  header_attr :date,  Time,   R | W | NODUP

  redefine_eql
  empty_able


  private

  def init
    @from = @by = @via = @with = @msgid = @for_ = nil
    @with = []
    @date = nil
  end

  Tag = %w( from by via with id for )

  def do_accept( stra )
    val = [ @from, @by, @via ] + @with + [ @msgid, @for_ ]
    val.compact!
    c   = [ @from ? 1 : 0, @by ? 1 : 0, @via ? 1 : 0,
            @with.size, @msgid ? 1 : 0, @for_ ? 1 : 0 ]

    i = 0
    c.each_with_index do |count, idx|
      label = Tag[idx]
      count.times do
        stra.spc unless i == 0
        stra.meta label
        stra.spc
        stra.meta val[i]
        i += 1
      end
    end

    stra.meta ';'
    if @date then
      stra.spc
      stra.meta ::TMail.time2str( @date )
    end
  end

end


# message-id type  ----------------------------------------------------

class MsgidH < StructH

  parse_on_rw

  def initialize( fname, fbody, strict )
    super
    @msgid = fbody.strip
  end

  header_attr :msgid, String, R | W
  
  redefine_eql
  empty_able


  private

  def init
    @msgid = nil
  end

  def do_accept( stra )
    stra.meta @msgid
  end

end


# encrypted type ------------------------------------------------------

class EncH < StructH

  parse_on_rw

  header_attr :encrypter, String, R | W
  header_attr :keyword,   String, R | W

  redefine_eql
  empty_able


  private

  def init
    @encrypter = nil
    @keyword = nil
  end

  def do_accept( stra )
    if @key then
      stra.meta @encrypter + ','
      stra.spc
      stra.meta @keyword
    else
      stra.meta @encrypter
    end
  end

end


# version type -----------------------------------------

class VersionH < StructH

  parse_on_rw

  header_attr :major, :Integer, R | W | NODUP
  header_attr :minor, :Integer, R | W | NODUP

  redefine_eql
  empty_able

  def version
    sprintf( '%d.%d', major, minor )
  end

  
  private

  def init
    @major = nil
    @minor = nil
  end

  def do_accept( stra )
    stra.meta "#{@major}.#{@minor}"
  end

end


# content type  ---------------------------------------


class CTypeH < StructH

  parse_on_create

  header_attr :main,   :String, R | W
  header_attr :sub,    :String, R | W
  header_attr :params, :Hash,   R | NODUP

  redefine_eql
  empty_able

  def []( key )
    params[key]
  end


  private

  def init
    @main = @sub = nil
    @params = {}
  end

  def do_accept( stra )
    stra.meta "#{@main}/#{@sub}"
    @params.each do |k,v|
      stra.meta ';'
      stra.spc
      stra.meta k
      stra.meta '='
      stra.value v
    end
  end

end


# encoding type  ----------------------------

class CEncodingH < StructH

  parse_on_rw

  header_attr :encoding, :String, R | W

  redefine_eql
  empty_able


  private

  def init
    @encoding = nil
  end

  def do_accept( stra )
    stra.meta @encoding
  end

end


# disposition type ----------------------------------

class CDispositionH < StructH

  parse_on_rw

  header_attr :disposition, :String, R | W
  header_attr :params,      :Hash,   R | NODUP

  def []( key )
    params[key]
  end

  redefine_eql
  empty_able
  

  private

  def init
    @disposition = nil
    @params = {}
  end

  def do_accept( stra )
    stra.meta @disposition
    @params.each do |k,v|
      stra.meta ';'
      stra.spc
      stra.meta k
      stra.meta '='
      stra.value v
    end
  end
    
end


# ---------------------------------------------------

class HeaderField   # backward definition

  STR2CLASS = {
    'date'                      => DateH,
    'resent-date'               => DateH,
    'received'                  => RecvH,
    'return-path'               => RetpathH,
    'sender'                    => SaddrH,
    'resent-sender'             => SaddrH,
    'to'                        => MaddrH,
    'cc'                        => MaddrH,
    'bcc'                       => MaddrH,
    'from'                      => MmboxH,
    'reply-to'                  => MaddrH,
    'resent-to'                 => MaddrH,
    'resent-cc'                 => MaddrH,
    'resent-bcc'                => MaddrH,
    'resent-from'               => MmboxH,
    'resent-reply-to'           => MaddrH,
    'message-id'                => MsgidH,
    'resent-message-id'         => MsgidH,
    'content-id'                => MsgidH,
    'in-reply-to'               => RefH,
    'references'                => RefH,
    'keywords'                  => KeyH,
    'encrypted'                 => EncH,
    'mime-version'              => VersionH,
    'content-type'              => CTypeH,
    'content-transfer-encoding' => CEncodingH,
    'content-disposition'       => CDispositionH,
    'subject'                   => StringH,
    'comments'                  => StringH,
    'content-description'       => StringH
  }

end


end   # module TMail
